# Copyright Red Hat
#
# This file is part of relval.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Adam Williamson <awilliam@redhat.com>
# Based on initial implementation by Josef Skladanka <jskladan@redhat.com>
# Josef's work incorporated code from Kamil Paral <kparal@redhat.com>
# Original in https://git.fedorahosted.org/cgit/fedora-qa.git/tree/stats

# this code is kinda irredeemable from these perspectives
# pylint: disable=too-many-locals,too-many-statements,too-many-branches,too-many-instance-attributes

"""This file contains functions that help implement the
'testcase-stats' sub-command of relval, which produces statistics
about release validation test coverage in HTML format.
"""

import datetime
import os
import re
import shutil
from collections import defaultdict, Counter, OrderedDict
from operator import attrgetter

from markupsafe import escape
from pkg_resources import resource_string, resource_filename

SANITY_SUB = re.compile(r"\W")


class Test:
    """An object representing information for one 'unique test' (see
    main() for definition). Counts total passes, fails and warns for
    the test across all composes as 'p', 'f' and 'w', and creates a
    'bitmap' of these totals as a dict keyed by compose, each entry
    being a list with two entries, the first being the results total
    and the second a string that is used to style the HTML output
    depending on if the results are all passes, all fails or a mix.
    Keeps a note of the most recent compose for which it was sent a
    result row: this relies on the result rows being sent in the
    correct order, this class does not know how to order the results
    it is sent.

    Stores lists of the wikitcms result objects for this test in
    three more dicts, one each for 'passes', 'fails' and warns', keyed
    by compose. As it extracts result objects from the result row, it
    notes the environments for which any results were provided, and
    divides this into the number of possible environments to produce
    a coverage percentage for each compose. This information is used
    to generate the detailed result pages.
    """

    def __init__(self):
        self.passes = defaultdict(list)
        self.warns = defaultdict(list)
        self.fails = defaultdict(list)
        self.coverage = defaultdict(int)
        self.composes = set()
        self.bitmap = {}
        self.last_tested = ""
        self.envs = set()
        self.resultenvs = set()
        self.last_tested_by_env = {}
        self.milestone = ""
        self.secid = 0

    def update(self, compose, row):
        """Passed a result row and the 2-tuple that identifies the
        compose it comes from. Parses the data from the result row as
        described in the class description. Assumes there will only
        ever be one result row per test instance per compose: if this
        is ever not true, the bitmap will represent only the results
        from the last row seen, and the coverage percentage will be
        inaccurate.
        """
        self.milestone = row.milestone
        self.composes.add(compose)
        self.secid = row.secid
        self.envs.update(list(row.results.keys()))
        rlists = list(row.results.items())
        pas = 0
        fail = 0
        warn = 0
        envs = 0
        for env, rlist in rlists:
            self.last_tested_by_env.setdefault(env, "nottested")
            if rlist:
                # these are all the envs for which we have real
                # results in any compose; used by the post-processor
                self.resultenvs.add(env)
                envs += 1
            for result in rlist:
                if result.status == "pass":
                    self.last_tested = compose
                    self.last_tested_by_env[env] = compose
                    self.passes[compose].append((env, result))
                    pas += 1
                if result.status == "fail":
                    self.last_tested = compose
                    self.last_tested_by_env[env] = compose
                    self.fails[compose].append((env, result))
                    fail += 1
                if result.status == "warn":
                    self.last_tested = compose
                    self.last_tested_by_env[env] = compose
                    self.warns[compose].append((env, result))
                    warn += 1
        self.coverage[compose] = envs * 100 // len(row.results)
        self._update_bitmap(compose, pas, fail, warn)

    def _update_bitmap(self, compose, pas, fail, warn):
        self.bitmap[compose] = [0, "nottested"]
        if pas + fail + warn > 0:
            bmap = self.bitmap[compose]
            if fail + warn == 0:  # just passed
                bmap[1] = "pass"
            elif pas + warn == 0:  # just failed
                bmap[1] = "fail"
            else:
                bmap[1] = "warn"
            bmap[0] += pas + fail + warn

    def merge_from(self, other):
        """Passed another Test object, this merges the contents of
        that Test into this one. Used by the post-processor to clean
        results.
        """
        for compose in other.composes:
            self.composes.add(compose)
            self.passes[compose] = other.passes[compose]
            self.fails[compose] = other.fails[compose]
            self.warns[compose] = other.warns[compose]
            self.coverage[compose] = other.coverage[compose]
            pas = len(other.passes[compose])
            fail = len(other.fails[compose])
            warn = len(other.warns[compose])
            self._update_bitmap(compose, pas, fail, warn)
        if not self.last_tested:
            self.last_tested = other.last_tested

    def get_page_contents(self, compose):
        """Generates the content for the detailed result pages for
        each 'unique test', which are linked from the summary page.
        """
        body = ""
        if compose not in self.composes:
            # There's a page for this compose and test type, but this
            # 'test instance' doesn't seem to exist in it (happens if
            # test got renamed or moved sections).
            body += "<b>Test not found for compose.</b>\n</p>\n"
            return body
        if self.coverage[compose] == 0:
            body += "<b>No results.</b>\n</p>\n"
            return body
        body += f"<b>Coverage:</b> {self.coverage[compose]}%<br/>"
        # 'gtr' lets us get the results for the status we're currently
        # working on. results for each status are stored as tuples of
        # the result environment and the result object itself.
        #
        #'cntr' counts the occurrences of each environment in the set
        # of results for the status; the following line reads out the
        # list of envs 'cntr' found and prints the name of each env
        # and the count of results that was found for it.
        for status in ("Passes", "Warns", "Fails"):
            body += f"<b>{status}: </b>"
            gtr = attrgetter(status.lower())
            cntr = Counter([env for (env, res) in gtr(self)[compose]])
            body += ", ".join([f"{escape(env)} - <b>{cntr[env]}</b>" for env in cntr])
            body += "\n<ul>"
            for _, result in gtr(self)[compose]:
                # we don't want to include the bug if it's not just a digit
                # (e.g. comment in the wrong place, or smth like 'bgo#714')
                rbugs = [bug for bug in result.bugs if bug.isdigit()]
                if rbugs:
                    # Add a list of bugs for this status, if any.
                    url = '{0}: <a href="https://bugzilla.redhat.com/show_bug.cgi?id={1}">{1}</a>'
                    body += "\n<li>" + "\n<li>".join(
                        [url.format(escape(result.user), escape(bug)) for bug in rbugs]
                    )
            body += "\n</ul>\n"
        return body


def get_bitmap_html(bitmap, allcomposes):
    """Renders the HTML for the little graphic that illustrates test
    coverage for a given 'unique test' in the final output. Uses the
    data from the Test() object's bitmap property, which is a dict of
    tuples keyed by compose, each tuple providing a result count and
    a string indicating whether any results are all passes, all fails,
    or a mix of the two; that string as a CSS class makes the dot(s)
    green, red or orange (or grey for 'nottested' if 0 results). If
    we believe the result row is entirely missing for a compose, an
    empty space will be rendered ('sym_nopage').
    """
    out = '<span style="letter-spacing:-0.5em;font-size:1.5em;">'
    # These are the little 'dot' symbols the coverage bitmap uses.
    sym_nopage = "&#10240;"
    sym_0 = "&#10244;"
    sym_1 = "&#10244;"
    sym_2 = "&#10246;"
    sym_3 = "&#10247;"
    for compose in allcomposes:
        try:
            (cnt, style) = bitmap[compose]
            out += f'<span class="{style}">'
            if cnt == 0:
                out += sym_0
            elif cnt < 3:
                out += sym_1
            elif cnt < 5:
                out += sym_2
            else:
                out += sym_3
            out += "</span>"
        except KeyError:
            # if that failed, it means this test did not exist for
            # this compose.
            out += sym_nopage
    out += "</span>"
    return out


def post_process(tests, allcomposes):
    """Crazy magic voodoo which cleans up the results after they've
    been turned into Test() objects but before we feed them to
    print_results_html(). Is passed the tests dict for a testtype,
    and the list of all composes for which any tests at all exist in
    the set of results being considered.

    The first half of the processor tries to merge tests which are
    'really' the same. First thing it does is remove lots of tests from
    consideration entirely. If the test case occurs exactly once in the
    rows for the test type, we don't currently do any merges that would
    be relevant to it, so we throw it away (we're not handling test
    case renames yet). It also throws out any test which appears in the
    page for every single compose; again, none of the merges we
    currently do would merge into such a test.

    Next, it tries to catch cases where tests were moved between
    sections or sections were renamed. It iterates over the cases that
    remain in its 'to be considered' list. For each one, it finds the
    other tests in the list which have the same test name + test case
    (and hence only differ in section name). It then filters those
    tests down to only the ones which do not include results for any of
    the same composes as the test being matched. It then reduces the
    'candidate' list once more, by including only tests which have all
    the environments (result columns) the test being matched has
    results for.

    Then it merges the test being matched with the first test in the
    remaining candidate list, if there still *is* one. That might sound
    a little rash, but in practice, it's pretty much always correct.
    There *are* cases where multiple candidates still remain after the
    final sort, but given how we organize the pages, the top candidate
    in the list is almost always the right one - because the list is
    ordered and we're comparing top down. Say a page has three sections
    with identical names, and each section contains all the same test
    cases and environments, and all three sections get renamed between
    two composes: this parser handles that correctly. When it hits the
    first section in the list and runs through all its checks, there
    will be three candidates remaining - but the top one will be the
    correct one, because essentially you have all six sections in order
    at this point, and the results from section 1 drop into section 4,
    then from section 2 into section 5, and section 3 into section 6.

    For an encore, it re-sorts the results by section. Using an ordered
    dict for tests as it's created keeps them in order so long as
    the pages are perfectly consistent, but when there's a difference
    between pages, the 'changed' tests just get stuck on the bottom.
    So what we do here is we find all the categories in the final
    result list, then we iterate over the whole list updating the
    'section id' for each section as we go. Each result row stores the
    'id' - the numerical index - of the section it was in at the time
    wikitcms ran into it. With this trick, what we wind up with is a
    dict of all the sections in the final results, with each one having
    the highest numerical index it ever had on any of the pages. This
    gives us a sort order for the sections which should usually reflect
    the order of the sections on the last page in the set of pages.

    Then we build a new copy of the tests dict, sorted by the test's
    section name's index from the dict we just created. That way all
    the results with the same section name stay together, and the
    sections themselves are ordered correctly.

    The other nice thing about all this crazy magic voodoo is if you
    are currently making the sign of the cross at the screen, it all
    drops out completely if you just don't call the post-processor;
    send tests straight to print_results_html() and you'll get the raw
    results with none of this processing voodoo.
    """
    testlist = list(tests.keys())
    cnt = Counter([testcase for (name, testcase, section) in testlist])
    print(f"Total unique tests: {len(testlist)}")

    # don't consider test cases that occur only once...
    testlist = [tst for tst in testlist if cnt[tst[1]] != 1]
    # ...or tests which exist for all composes
    testlist = [tst for tst in testlist if tests[tst].composes != set(allcomposes)]
    print(
        "Merge candidates (tests with more than one occurrence, that do not "
        f"occur in all composes): {len(testlist)}"
    )

    # next iteration: now we're only looking at merge candidate cases
    for i, test in enumerate(testlist):
        # we only ever merge "down" - older results into newer ones
        cands = [c for c in testlist[i + 1 :] if (c[0], c[1]) == (test[0], test[1])]
        # print("Candidates w/ same name: " + str(len(cands)))
        cands = [c for c in cands if tests[test].composes.isdisjoint(tests[c].composes)]
        # print("Candidates without overlapping composes: " + str(len(cands)))
        cands = [c for c in cands if tests[test].resultenvs <= tests[c].envs]
        # print("Candidates w/ required envs: " + str(len(cands)))
        try:
            tests[cands[0]].merge_from(tests[test])
            tests.pop(test)
            tmpl = (
                "Merged test case: {0} with name: {1} from section: {2} "
                "into test case: {3} with name: {4} from section: {5}"
            )
            print(tmpl.format(test[1], test[0], test[2], cands[0][1], cands[0][0], cands[0][2]))
        except IndexError:
            # well I never commented this when I wrote it, but I'm
            # guessing this is for when cands is empty?
            pass

    # this finds all section names and stores the highest index found for each
    # in a dict keyed on the section name
    secids = {}
    secset = {section for (_, _, section) in tests.keys()}
    for section in secset:
        secids[section] = int(
            max(test.secid for (_, _, sec), test in tests.items() if sec == section)
        )

    # now we build the new tests dict sorted by the index for the section
    # name from the dict we just built
    tests = OrderedDict(sorted(iter(tests.items()), key=lambda x: secids[x[0][2]]))
    return tests


def print_results_html(testtype, pages, allcomposes, tests, outdir):
    """Produces the summary page for each 'testtype' (Installation,
    Base etc). Holy mother of God, this is awful. We need to get some
    Jinja up in here stat.
    """
    # template-y bits used to construct the page HTML
    col_names = ("Testcase URL", "Release Level", "Last in", "Coverage", "Details")
    trs = []
    thead = (
        '<h4>{0}</h4>\n\t<table id="testcases-table-{0}"'
        'class="tablesorter" border="0" cellpadding="0" cellspacing="1">'
        "\n\t\t<thead>\n\t\t\t<tr>"
    )
    thead += "\n".join([f"<th>{col}</th>" for col in col_names])
    thead += "\n\t\t\t</tr>\n\t\t</thead>\n\t\t<tbody>"

    tend = "\t\t</tbody>\n\t</table>"

    row_pre = (
        '<td style="padding=0"><a href="https://fedoraproject.org/'
        'wiki/{testcase}">{testcase}{name}</a></td>'
    )

    row_main = '<td style="padding=0">{}</td>\n' * (len(col_names) - 2)

    row_post = '<td style="padding=0"><a href="{link}.html">Details</a></td>\n'

    lastsec = ""
    firsttest = True

    # this is where we actually start doin' stuff
    os.makedirs(os.path.join(outdir, SANITY_SUB.sub("_", testtype)))
    for (name, testcase, section), test in tests.items():
        # Sanitize inputs, Bobby
        name = escape(name)
        testcase = escape(testcase)
        section = escape(section)
        # If this is the very first row in the page, create the first
        # table header.
        if firsttest:
            trs.append(thead.format(section))
            firsttest = False
        # Otherwise, start a new table if we've changed section.
        elif section != lastsec:
            trs.append(tend)
            trs.append(thead.format(section))
        # Update the variable that records the last section we saw.
        lastsec = section
        # We use slightly different formats for both the detail page
        # filenames and the summary page visible names for tests whose
        # 'name' differs from their 'testcase'. We want to display and
        # use both attributes when they differ, but just use one when
        # they're the same.
        if testcase != name:
            dispname = " - " + name
            linkname = "-" + name
        else:
            dispname = ""
            linkname = ""
        # If there's a section for the test case, include it in the
        # page URL in a nice way.
        if section:
            secturl = "-" + section
        else:
            secturl = ""
        trow = row_pre.format(testcase=testcase, name=dispname)
        trow += row_main.format(
            escape(test.milestone),
            escape(test.last_tested),
            get_bitmap_html(test.bitmap, allcomposes),
        )
        trow += row_post.format(
            link=SANITY_SUB.sub("_", testtype)
            + "/"
            + SANITY_SUB.sub("_", testcase)
            + SANITY_SUB.sub("_", linkname)
            + SANITY_SUB.sub("_", secturl)
        )
        trs.append(f"<tr>{trow}</tr>")

        # write detailed result page for this test
        template = resource_string(__name__, "html/testcase.html").decode("utf-8")
        body = ""
        for page in pages:
            body += (
                f'<p><strong><a href="https://fedoraproject.org/wiki/{escape(page.name)}">'
                f"{escape(page.name)}</a></strong><br />\n"
            )
            body += test.get_page_contents(escape(page.shortver))
            body += "</p>"
        fname = os.path.join(
            outdir,
            SANITY_SUB.sub("_", testtype),
            SANITY_SUB.sub("_", testcase)
            + SANITY_SUB.sub("_", linkname)
            + SANITY_SUB.sub("_", secturl)
            + ".html",
        )
        try:
            if testcase != name:
                outname = " - " + name
            else:
                outname = ""
            with open(fname, "w", encoding="utf-8") as fout:
                fout.write(
                    template.format(
                        timestamp=datetime.datetime.utcnow(),
                        testcase=testcase,
                        name=outname,
                        body=body,
                    )
                )
        except IOError:
            pass

    trs.append(tend)
    body = "\n".join(trs)
    # create the Overview page
    outfile = f"{os.path.join(outdir, SANITY_SUB.sub('_', testtype))}.html"
    print(f"Creating HTML output: {outfile}")
    template = resource_string(__name__, "html/index.html").decode("utf-8")
    with open(outfile, "w", encoding="utf-8") as fout:
        fout.write(template % {"timestamp": datetime.datetime.utcnow(), "body": body})


def get_results_data(allcomposes, tests):
    """Returns a dictionary containing the testcase-stats, to be used
    for the JSON formatted output."""
    data = {}
    for (name, testcase, section), test in tests.items():
        # Look ma, no hands!
        data.setdefault(section, [])
        if testcase != name:
            dispname = name
        else:
            dispname = None
        bitmap = {c: [0, "n/a"] for c in allcomposes}
        bitmap.update(test.bitmap)

        trow = {
            "testcase_url": f"https://fedoraproject.org/wiki/{testcase}",
            "testcase_name": testcase,
            "testcase_display_name": dispname,
            "milestone": test.milestone,
            "last_tested": test.last_tested,
            "last_tested_by_env": test.last_tested_by_env,
            "bitmap": bitmap,
            "inactive": bitmap[allcomposes[-1]] == "n/a",
        }
        data[section].append(trow)
    return data


def print_summary_html(outdir, allpages):
    """Despite the name, this writes the top-level index.html linking
    to the page for each test type.
    """
    # copy javascript
    shutil.copytree(resource_filename(__name__, "html/js"), os.path.join(outdir, "js"))

    # Create index page
    with open(os.path.join(outdir, "index.html"), "w", encoding="utf-8") as fout:
        fout.write("<html><body>\n")
        fout.write(f"<p>Created at: {datetime.datetime.utcnow()} (UTC)</p>")
        for testtype in allpages.keys():
            fout.write(f'<a href="{SANITY_SUB.sub("_", testtype)}.html">{testtype}</a><br />')
        fout.write("\n</body></html>")
